require 'erb'
require 'optparse'
require 'ostruct'
require 'yaml'

NO_CHANGE = Object.new

class Splice
  attr_accessor :array

  def initialize(array)
    self.array = array
  end
end

class Processor
  attr_writer :file_content
  attr_accessor :source_file

  def initialize(source_file:, file_content: nil)
    self.source_file = source_file
    self.file_content = file_content
  end

  def hash
    hash = file_content
    process! hash
    hash
  end

  private

  def file_content
    @file_content ||= YAML.load_file(source_file)
  end

  def relative_path(path)
    File.expand_path(path, File.dirname(source_file))
  end

  def maybe_processed!(o)
    data = process!(o)
    if data == NO_CHANGE
      [o, false]
    else
      [data, true]
    end
  end

  def process!(o)
    if o.is_a?(Hash)
      if o.has_key?("__dy__")
        process_dy!(o)
      else
        process_hash!(o)
      end
    elsif o.is_a?(Array)
      process_array!(o)
    else
      NO_CHANGE
    end
  end

  def process_dy!(dy)
    case dy["__dy__"]
    when "escape" then o["value"]
    when "text_file"
      read_file(relative_path(dy["file"]), erb: dy["erb"], args: dy["args"])
    when "data_file"
      file = relative_path(dy["file"])
      content = read_file(file, erb: dy["erb"], args: dy["args"])
      data = YAML.load(content)
      data = Processor.new(source_file: file).hash
      dy["splice"] ? Splice.new(data) : data
    else
      raise "Unknown __dy__ #{dy["__dy__"].inspect}"
    end
  end

  def process_hash!(hash)
    changed = false
    hash.each do |key, value|
      hash[key], processed = maybe_processed!(value)
      changed ||= processed
    end
    changed ? hash : NO_CHANGE
  end

  def process_array!(array)
    any_splice = false
    changed = false
    array.each_with_index do |value, index|
      array[index], processed = maybe_processed!(value)
      any_splice ||= array[index].is_a?(Splice)
      changed ||= processed
    end

    splice_array!(array) if any_splice

    changed ? array : NO_CHANGE
  end
end

def splice_array!(array)
  index = 0
  while index < array.length do
    value = array[index]
    if value.is_a?(Splice)
      raise "Cannot splice except arrays" unless value.array.is_a?(Array)
      array[index, 1] = value.array
      index += value.array.length
    else
      index += 1
    end
  end
end

def read_file(filename, erb: false, args: {})
  content = File.read(filename)
  if erb
    renderer = ERB.new(content)
    renderer.result(OpenStruct.new(args).instance_eval { binding })
  else
    content
  end
end

def read_arguments
  options = {}
  OptionParser.new do |opts|
    opts.banner = "Usage: test.rb [options]"

    opts.on("-pPREFIX", "--prefix=PREFIX", "Prefix for the yaml output") do |v|
      options[:prefix] = v
    end
  end.parse!

  src, dst, *more = ARGV
  if src.nil? || dst.nil? || !more.empty?
    raise "Need two arguments"
  end

  [src, dst, options]
end

def main
  src, dst, options = read_arguments
  hash = Processor.new(source_file: src).hash
  File.write(dst, "#{options[:prefix] + "\n" if options[:prefix]}#{hash.to_yaml}")
end

main
